
#import python modules
import numpy as np
import time
import atomwrite

#import compiled fortran library
import fplib
import sys
import os

#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis

"""========== GLOBAL VARIABLES =========="""
Cut = 3 # cut off distance of LJ potential
Tort_Coe=0.3  # coefficient to tune pore tortuosity
PoreShape=1   # parameter to build pore shape [0: hexagonal, 1: cubic 2: gyroid]
Rho=0.0015 # density corresponding to 2 atm of pressure
Lx=46 # unit cell x dimension
W=14 # pore width
Lz=30 #unit cell z dimension

#pore-pore distance ~ 10 nm
#Pore diameter ~ 6 nm

np.random.seed(342324)

#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.


### functions for implicit pore wall

## hexagonal
def ParPore_B(x):
    return -0.5*W
def ParPore_C(x):
    return 0.5*W

def ParPore_B_der(x):
    return 0
def ParPore_C_der(x):
    return 0

## cubic
def CubicPore_B(x):
    return W*(Tort_Coe*np.sin(4.*np.pi*x/Lx-np.pi)-1./2.)
def CubicPore_C(x):       
    return W*(Tort_Coe*np.sin(4.*np.pi*x/Lx)+1./2.)

def CubicPore_B_der(x):
    return W*Tort_Coe*4.*np.pi/Lx*np.cos(4.*np.pi*x/Lx)
def CubicPore_C_der(x):
    
    return W*Tort_Coe*4.*np.pi/Lx*np.cos(4.*np.pi*x/Lx)

## gyroid
def GyPore_B(x):
    return W*(Tort_Coe*np.sin(4.*np.pi*(x/Lx))-1./2.)
def GyPore_C(x):
    return W*(Tort_Coe*np.sin(4.*np.pi*(x/Lx))+1./2.)

def GyPore_B_der(x):
    return W*Tort_Coe*4.*np.pi/Lx*np.cos(4.*np.pi*(x/Lx)-np.pi)
def GyPore_C_der(x):
    return W*Tort_Coe*4.*np.pi/Lx*np.cos(4.*np.pi*(x/Lx))




def CPore_B(x,y):
    return W*(Tort_Coe*np.sin(4.*np.pi*(x/L)-np.pi)-1./2.)
    
def CPore_C(x,y):
    return W*(Tort_Coe*np.sin(4.*np.pi*(x/L))+1./2.)


def Pore_B(Posi):
    global PoreShape
    Poresurface=[ParPore_B(Posi[0]), CubicPore_B(Posi[0]), GyPore_B(Posi[0])]
    return Poresurface[PoreShape]

def Pore_C(Posi):
    global PoreShape
    Poresurface=[ParPore_C(Posi[0]), CubicPore_C(Posi[0]), GyPore_C(Posi[0])]
    return Poresurface[PoreShape]
    
    
def Pos_W(Posi):
    global PoreShape
    Pos_WB=Posi.copy()
    Pos_WC=Posi.copy()
    #print "Y values", YPos_W(Posi)
    Pos_WB[1], Pos_WC[1]=YPos_W(Posi)   
    #print Posi, Pos_WB, Pos_WC
    return Pos_WB, Pos_WC

    
def squareD(PosA, PosB):
    return sum((PosA-PosB)*(PosA-PosB))
    

def YPos_W(Posi):
    global PoreShape
    Bsurface=[ParPore_B(Posi[0]), CubicPore_B(Posi[0]), GyPore_B(Posi[0])]
    Csurface=[ParPore_C(Posi[0]), CubicPore_C(Posi[0]), GyPore_C(Posi[0])]
    #print Bsurface[PoreShape], Csurface[PoreShape]
    return Bsurface[PoreShape], Csurface[PoreShape]
    
def YPos_W_der(Posi):
    global PoreShape
    Bsurface_der=[ParPore_B_der(Posi[0]), CubicPore_B_der(Posi[0]), GyPore_B_der(Posi[0])]
    Csurface_der=[ParPore_C_der(Posi[0]), CubicPore_C_der(Posi[0]), GyPore_C_der(Posi[0])]
    return Bsurface_der[PoreShape], Csurface_der[PoreShape]
    
def GetPoreW(Pos):
    global PoreShape
    PoreW_C=Pos.copy()
    PoreW_B=Pos.copy()
    i=0
    while i < len(Pos):
        PoreW_B[i,1], PoreW_C[i,1]=YPos_W(Pos[i])
        i+=1
    return PoreW_B, PoreW_C
    
def GetC_poreW(Pos):
    global PoreShape
    
    PosWB=np.zeros_like(Pos)
    PosWC=np.zeros_like(Pos)
    i=0
    while i < len(Pos):
        PosWB[i], PosWC[i]=Closest_wall(Pos[i])
        #print "done for particle ", i
        i+=1
    
    return PosWB, PosWC



def Closest_wall(Posi, dx=0.01, EFracTol=1.e-1, 
               Accel = 10, MaxInc = 10., MaxIter = 10000):
    """
Find a point on the pore wall that is closest to the particle    
    
"""
   
    
    #identify global variables
    global Lx, W, Lz, Cut, PoreShape

    #start the iteration counter
    idxBC=[0,1]
    PosiWC=np.zeros_like(Posi)
    PosiWB=np.zeros_like(Posi)
    PosiW=[PosiWC, PosiWB]
    for idx in idxBC:
        NormDir = np.array([1,0,0], float)        
        
        if squareD(Posi,Pos_W(Posi)[idx]) < 400:
            
                    
            Iter = 0
    #find the normalized direction    
        
    #take the first two steps and compute energies    
            Dists = [0., dx]
            #PosiW[idx]=YPos_W(Posi)[idx]
            direct=2*YPos_W(Posi)[idx]-Posi[1]*YPos_W_der(Posi)[idx]
            NormDir=-1*NormDir*direct/np.sqrt(direct*direct)
        #print "NormDir", NormDir
            R2s = [squareD(Posi,Pos_W(Posi + NormDir * x)[idx]) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
            while R2s[1] > R2s[0]:
                Iter += 1
                dx = dx * 0.5
                Dists[1] = dx
                R2s[1] = squareD(Posi,Pos_W(Posi + NormDir * dx)[idx])
            #print "Iter, step1" , Iter
    #find a third point
            Dists = Dists + [2. * dx]
            R2s = R2s + [squareD(Posi,Pos_W(Posi + NormDir *2* dx)[idx])]
        #print R2s, "========))))"
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
            while R2s[2] < R2s[1]:
                Iter += 1
            
        #find a fourth point and evaluate energy
                Dists = Dists + [Dists[-1] + dx]
        
                R2s = R2s + [squareD(Posi,Pos_W(Posi + NormDir*Dists[-1])[idx])]
            

        #check if we increased too much in energy; if so, back off
         #   print R2s, "======="
                if (R2s[3] - R2s[0]) > MaxInc * (R2s[0] - R2s[2]):
                    R2s = R2s[:3]
                    Dists = Dists[:3]
                    dx = dx * 0.5
                else:
            #shift all of the points over
                    R2s = R2s[-3:]
                    Dists = Dists[-3:]
                    dx = dx * Accel
            
    #if (R2s[1] > 100) :     
    #    return np.array([0, 2*Cut, 2*Cut], float), Cut*8
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
            OldR23 = W
            
            while True:
                Iter += 1
                if Iter > MaxIter:
                    print "Warning: maximum number of iterations reached in closest wall search."
                    break
            
        #store distances for ease of code-reading
                d0, d1, d2 = Dists
                R20, R21, R22 = R2s

        #use a parobolic approximation to estimate the location
        #of the minimum
                d10 = d0 - d1
                d12 = d2 - d1
                Num = d12*d12*(R20-R21) - d10*d10*(R22-R21)
                Dem = d12*(R20-R21) - d10*(R22-R21)
                if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
                    d3 = 0
                else:
            #location of parabolic minimum
                    d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        
                R23 = squareD(Posi,Pos_W(Posi + NormDir *d3)[idx])
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
                if d3 < d0 or d3 > d2 or R23 > R20 or R23 > R21 or R23 > R22:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
                    if abs(d2 - d1) > abs(d0 - d1):
                        d3 = 0.5 * (d2 + d1)
                    else:
                        d3 = 0.5 * (d0 + d1)
            
                    R23 = squareD(Posi,Pos_W(Posi + NormDir *d3)[idx])
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
                if d3 < d1:
                    if R23 < R21:
                #get rid of point 2
                        Dists, R2s = [d0, d3, d1], [R20, R23, R21]
                    else:
                #get rid of point 0
                        Dists, R2s = [d3, d1, d2], [R23, R21, R22]
                else:
                    if R23 < R21:
                #get rid of point 0
                        Dists, R2s = [d1, d3, d2], [R21, R23, R22]
                    else:
                #get rid of point 2
                        Dists, R2s = [d0, d1, d3], [R20, R21, R23]
                
        #check how much we've changed
                if abs(OldR23 - R23) < EFracTol * abs(R23):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
                    break
                OldR23 = R23

    #return the position array at the minimum (point 1)        
            PosiW[idx]=Pos_W(Posi + NormDir *Dists[1])[idx]    
            
        else:
            PosiW[idx]=np.array([0, (float(idx)-0.5)/0.5*1000, 0], float)
            #print "Dists", Dists[1]
            #print "Iter=", Iter        
    return PosiW
         

def InitPositions(N, Lx, Ly, W):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    Coeff=1.0
    while True :
        Coeff=Coeff-0.02
    #make the position array
        Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
        VN=N*2
        LatSpac=(Lx*2*W*Lz/VN)**(1./3.)*Coeff
        NLatx=int(Lx/LatSpac)
        
        NLaty=int(2*W/LatSpac)
        NLatz=int(Lz/LatSpac)
    #LatSpacex=Lx/float(NLatx)
    #LatSpacey=Ly/float(NLatx)
    #LatSpacew=2*W/float(NLatx)
        
#        NLat = int((VN)**(1./3.) + 1.)
#    LatSpac = L / NLat
    #make an array of lattice sites
        rx = (LatSpac * (np.arange(NLatx, dtype=float)+0.5) - 0.5*Lx)
        ry = (LatSpac * (np.arange(NLaty, dtype=float)+0.5) - 0.5*2*W)
        rz = (LatSpac * (np.arange(NLatz, dtype=float)+0.5) - 0.5*Lz)
    
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
        i = 0
        for y in ry:
            for z in rz:
                for x in rx:
                    Posi=np.array([x,y,z], float)
                    if (YPos_W(Posi)[0]+1. < y < YPos_W(Posi)[1]-1.):
                        Pos[i] = Posi.copy()
                        #print YPos_W(Posi)[0]+1.,  y, YPos_W(Posi)[1]-1.
                    #    print Pos[i], Posi, "\n"
                    
                #add a random offset to help initial minimization
                    #Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5) #0.1???
                    #Pos[i] = Pos[i] + Offset
                        i += 1
                #if done placing atoms, return
                    if i >= N:
                        #print Pos
                        return Pos

        print "cannot reach to N. Adjusting the LatSpac"
        print "i, Coeff,", i, Coeff
    return Pos



def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature using Berendsen Thermostat
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * 32.*np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    #VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    VScale = np.sqrt(1+1./100.*(1.5*len(Vel)*T/KE-1))
    
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, L, Cut
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = fplib.calcenergyforces(Pos, PosWB, PosWC, Lx, W, Lz, Cut, Forces)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return 32.*np.sum(Vel * Vel) / (3. * len(Vel))




import matplotlib
import matplotlib.pyplot as plt
import random

N=int(Lx*W*Lz*Rho)
T=6.8
dt=0.001
dx=0.01
EFracTolLS=1.e-8
EFracTolCG=1.e-10
RescaleFreq=20
NStepsEquil1=10000 # 10000
NStepsEquil2=10000 # 10000
NStepsProd=60000  #60000
PdbFreq=400 #200
MSDFreq=200 #100
DisFreq=5000 # 5000
Ncalc=4 #4
m=0
dxbin=.5
L=Lx
Plot=False
DisCollect=True
print "N=", N, "Production=", NStepsProd
Dist=[]
Distlabel=[]


while m<Ncalc:
    seed=random.randrange(1,100000)
    np.random.seed(seed)
    Names=[] 
    Pos=InitPositions(N, Lx, W, Lz)
    i=0
    while i < len(Pos):
        Names.append("H")
        i+=1

    if Plot:
        t = np.arange(-1*Lx, Lx, 0.2)
        y1=[YPos_W(np.array([x,1,0], float))[0] for x in t]
        y2=[YPos_W(np.array([x,1,0], float))[1] for x in t]
        
            
        Gline=[1,1,1,1]
        plt.figure(num=1, figsize=(8,6), facecolor='w', edgecolor='k')    
        plt.subplot(111)
        plt.plot(t, y1, 'k', t, y2, 'k')
        plt.plot(Pos[:,0], Pos[:,1], 'b.')
        plt.axis([-0.5*Lx, 0.5*Lx, -1*W, W])
        plt.show()


    if m==0:
        WPosx=np.arange(Lx, dtype=float)
        WPosz=np.arange(Lx, dtype=float)
        
        WallSurfaceB=[]
        WallSurfaceC=[]
        for x in WPosx:
            for z in WPosz:
                WallSurfaceB.append([x, YPos_W(np.array([x,0,0], float))[0], z])
                WallSurfaceC.append([x, YPos_W(np.array([x,0,0], float))[1], z])
                Names.append("C")
                Names.append("C")
                
    
        Pdb=atomwrite.pdbfile("anim_S%i_Tort_Coef%.2f_m%i.pdb"%(PoreShape, Tort_Coe, m), Lx, W, Lz, AtomNames=Names)

        WallSurfaceB=np.array(WallSurfaceB)
        WallSurfaceC=np.array(WallSurfaceC)

    PosWB, PosWC=GetC_poreW(Pos)
    Forces = np.zeros_like(Pos)
    PE, Forces=fplib.calcenergyforces(Pos, PosWB, PosWC, Lx, W, Lz, Cut, Forces)
    Vel=InitVelocities(N, T)
    Accel=InitAccel(Pos)    




    i=0
    print "start eq1."

    while i<NStepsEquil1 :
        # Period 1, run MD with velocity rescaling
        Pos, Vel, Accel, KE, PE=fplib.vvintegrate(Pos, PosWB, PosWC, Vel, Accel, Lx, W, Lz, Cut,
                                               dt)
        PosWB, PosWC=GetC_poreW(Pos)
    
        if i % RescaleFreq ==0:
            Vel=RescaleVelocities(Vel, T)
        
        i+=1        
    dataE=[]    
    i=0
    print "start eq2."
    while i<NStepsEquil2 :
        # Period 1, run MD with velocity rescaling
        Pos, Vel, Accel, KE, PE=fplib.vvintegrate(Pos, PosWB, PosWC, Vel, Accel, Lx, W, Lz, Cut,
                                               dt)
        PosWB, PosWC=GetC_poreW(Pos)
    
        if i % RescaleFreq ==0:
            Vel=RescaleVelocities(Vel, T)
        i+=1        
        dataE.append(KE+PE)

    dataE=np.array(dataE)
    # Calculate average total energy in period 2    
    Eave=dataE.mean()
    # velocity rescaling to give target total Energy
    Vel=RescaleVelocities(Vel, (Eave-PE)/(1.5*len(Pos)))
    # save the current position to compute mean-square displacement
    Pos0=Pos.copy()
    #dPos=np.zeros_like(Pos, float)
    i=0
    DataMSD=[]
    DataDis=[]
    DataDislabel=[]
    fresult=file("Finalproject_S%i_T%.2f_m%i.txt"%(PoreShape, Tort_Coe, m),"a")
    #fresult2=file("Finalproject_S%i_T%f_m%i_distribution.txt"%(PoreShape, Tort_Coe, m),"a")
    print "start production"
    while i<NStepsProd :
        # Period 3, production time without velocity rescaling
        Pos, Vel, Accel, KE, PE=fplib.vvintegrate(Pos, PosWB, PosWC, Vel, Accel, Lx, W, Lz, Cut,
                                               dt)
        PosWB, PosWC=GetC_poreW(Pos)
        if m==0:
            ## save position on Pdb file for moive
            if i%PdbFreq==0:
            #PosP=Pos.copy()
                PdbPos=np.concatenate((Pos, WallSurfaceB, WallSurfaceC), axis=0)
#        print PdbPos.shape, Names.shape
        #print len(Names)
        #print len(PdbPos)
                Pdb.write(PdbPos)
        #print "pdb record"
       

        if UseVisual:
            if atomvis.Initialized:
            #update the positions
                atomvis.Update(Pos)
            else:
            #initialize the visualization window
                atomvis.Init(Pos)
        if i%MSDFreq==0:
            MSD=sum((Pos-Pos0)*(Pos-Pos0))/len(Pos)
    

        # save mean-square displacement
            DataMSD.append([PoreShape, W, i*dt,MSD[0], MSD[1], MSD[2]])
            
            fresult.write("%2i%6.4f%2i%10.5f%10.4f%10.4f%10.4f\n" %(PoreShape, Tort_Coe, m, i*dt,MSD[0], MSD[1], MSD[2]))
            

        if i%DisFreq==0 and DisCollect:
            disPos=Pos-Pos0
            DataDis.append(disPos)
            #print len(disPos), "len_disPos"
            DataDislabel.append([PoreShape, W, Tort_Coe, m, i*dt])
            #fresult2.write("%2i%4f%2i%10.5f%10.4f%10.4f%10.4f\n" %(PoreShape, Tort_Coe, m, i*dt, MSD[0], MSD[1], MSD[2]))
            
        
        i+=1    
    print "Calculation complete"
    if m==0:
        Pdb.close()
    fresult.write("TER\n")
    fresult.close()
        
    DataMSD=np.array(DataMSD)
    if DisCollect:
        Dist.append(DataDis)
        Distlabel.append(DataDislabel)
    print m, "m"        
    m +=1

## Save data for particle distribution fuction

if DisCollect:
    fresult2=file("Finalproject_S%i_T%.2f__distribution.txt"%(PoreShape, Tort_Coe),"a")
    Dist=np.array(Dist)
    Distlabel=np.array(Distlabel)
    Disfn=[]
    print Dist[:,:,:,0]
    print "max", Dist[:,:,:,0].max()
    print "min", -1*Dist[:,:,:,0].min()
    Lt=max([Dist[:,:,:,0].max(), -1*Dist[:,:,:,0].min()])
    binset=[dxbin*((k-.5)-(int(Lt/dxbin)+1)) for k in range((int(Lt/dxbin)+1)*2+2)]
    idx=0
    n_t=len(Dist[0])
    print n_t, "n_t"
    while idx < n_t:
        Disfn.append(np.zeros((len(binset)-1), int))
        idx +=1
    Disfn=np.array(Disfn)
    
    #print np.shape(Dist), "len(Dist)"
    #print np.shape(Dist[1])
    #print np.shape(Dist[1,1])
    #print np.shape(Dist[1,1,1])
    for mPos in Dist:
        idx=0
        print len(mPos), "len_mPos"
        while idx < n_t:
            for iPos in mPos[idx]:
                #print np.shape(iPos)                
                mm=0
                k=0
                
                print iPos[0]
                while k<len(binset)-1:
                    if (iPos[0] >= binset[k]) and (iPos[0]<binset[k+1]):
                        Disfn[idx, k]+=1
                        mm+=1
                        #print mm, k, Disfn[idx,k]
                    k+=1
            idx+=1
    idx=0
    print sum(sum(Disfn)), "sum_sum_disfn"
    
    #print Disfn.shape(), "shape"
    
    while idx < len(mPos):
        fresult2.write("%10f\n" %Distlabel[0,idx,4])
        fresult2.write("%5s%10s%10s%12s\n" %('nbin','lowlimit', 'highlimit', 'population'))
        j=0
        while j<len(binset)-1:
            #print "disfn"
            #print j+1, binset[j], binset[j+1], Disfn[idx,j]
            fresult2.write("%5i%10f%10f%12f\n" %(j+1, binset[j], binset[j+1], Disfn[idx,j]))
            j+=1
        fresult2.write("\n")
        idx +=1
    fresult2.write("TER\n")
    fresult2.close()
    
    #print binset            
    #print Disfn
    
                
        
    